/**
  *  \file client/dialogs/prefdlg.cc
  *  \brief Preferences dialog
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/prefdlg.h"
#include "client/chart.h"
#include "client/dialogs/chartconfig.h"
#include "client/markerconfig.h"
#include "client/widgets/configedit.h"
#include "client/widgets/configstorage.h"
#include "client/widgets/configview.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/lock.h"
#include "game/pref.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/cardgroup.h"
#include "ui/widgets/cardtab.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/radiobutton.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/tree.h"
#include "util/nls.h"
#include "client/dialogs/labeldlg.h"

namespace {
    class WConfigBackupEditor : public WConfigEditor::Editor {
     public:
        WConfigBackupEditor(Config& cfg);
        string_t toString(string_t arg);
        void edit(string_t arg);

     private:
        Config& cfg;
    };

    class WConfigBackupEditorWindow : public UIWindow {
     public:
        WConfigBackupEditorWindow(ConfigStringOption& opt, string_t def);
     private:
        ConfigStringOption& opt;
        string_t def;
        ObjectHolder h;
        Observable<int> selection;
        UIInputLine* input;

        void init();
        void onMove();
        void onOK();
    };

    class WConfigStarchartEditor : public WConfigEditor::Editor {
     public:
        string_t toString(string_t arg);
        void edit(string_t arg);
    };

    class WConfigBooleanEditor : public WConfigEditor::Editor {
     public:
        WConfigBooleanEditor(Config& cfg);
        string_t toString(string_t arg);
        void edit(string_t arg);

     private:
        Config& cfg;
    };

    class WConfigLabelEditor : public WConfigEditor::Editor {
     public:
        string_t toString(string_t arg);
        void edit(string_t arg);
    };

    enum EditorType {
        NoEditor,
        DefaultEditor,
        BackupEditor,
        StarchartEditor,
        BoolEditor,
        LabelEditor
    };

    /** Preferences dialog. */
    class WPreferencesDialog : public UIWindow {
     public:
        WPreferencesDialog(Config& cfg);

     private:
        void init();
        void onMove();

        Config& cfg;

        ObjectHolder h;
        UITreeWidget* settings;
        WConfigView* details;
        WConfigEditor* editor;
        WConfigStorageControl* storage;

        ptr_vector<WConfigEditor::Editor> editors;
    };

    struct ConfigItem {
        const char* name;
        int         level;
        const char* info;
        EditorType  editor;
        const char* arg;
    };
}

static const ConfigItem config_tree[] = {
    { N_("Maketurn"),                            0, "", NoEditor,         "" },
    {   N_("Backup..."),                         1, "", BackupEditor,     "Backup.Turn,%d/backups/turn%p.%t" },
    {   N_("Upload target..."),                  1, "", DefaultEditor,    "Maketurn.Target" },
    //   TRN file format
    //   Use all message files
    //   Non-Player messages
    //   PHost identification
    //   Wait after each race
    { N_("Unpack"),                              0, "", NoEditor,         "" },
    {   N_("Data file format"),                  1, "", BoolEditor,       "Unpack.Format" },
    //   Race name changes
    //   Convert text files in UTILx.DAT
    {   N_("Create TARGETx.EXT files"),          1, "", BoolEditor,       "Unpack.TargetExt" },
    //   Source directory...
    //   Backup directory...
    {   N_("Backup..."),                         1, "", BackupEditor,     "Backup.Result,%d/backups/result%p.%t" },
    {   N_("Error correction"),                  1, "", BoolEditor,       "Unpack.FixErrors" },
    //   Wait after each race
    { N_("Program Interoperation"),              0, "", NoEditor,         "" },
    //   Winplan directory...
    //   Decompile PHost commands
    //   Save PHost commands in messages
    //   Update HCONFIG.HST
    {   N_("Rewrap incoming messages"),          1, "", BoolEditor,       "Messages.RewrapInbox" },
    //   Host 3.22.031 Torp Safety
    { N_("Preferences"),                         0, "", NoEditor,         "" },
    {   N_("User Interface"),                    1, "", NoEditor,         "" },
    //     Confirm exit from race screen
    //     Expensive graphical effects
    //     Display options...
    //     Tip of the day
    //     Thousands-separator
    //     Use clans everywhere
    //     Cache ship lists
    {     N_("Waypoint tied to scanner"),        2, "", BoolEditor,       "Scanner.TieWaypoint" },
    //     Scanner/Waypoint independant
    //     Display Auto Waypoints in map
    //   VCR...
    {   N_("Starchart"),                         1, "", NoEditor,         "" },
    {     N_("General Settings..."),             2, "", StarchartEditor,  "" },
    {     N_("Object Labels..."),                2, "", LabelEditor,      "" },
    {   N_("Collapse old messages"),             1, "", BoolEditor,       "Messages.CollapseOld" },
    {   N_("Automatically lock at warp wells"),  1, "", BoolEditor,       "Chart.Scanner.WarpWells" },
    {   N_("Taxation predicts relative growth"), 1, "", BoolEditor,       "Tax.PredictRelative" },
    {   N_("Instant battle result"),             1, "", BoolEditor,       "VCR.InstantResult" },
    //   Guess ship positions
    //   Starcharts file (CHARTx.CC)
    //   Auto load...
    //   If UTILx.DAT available...
    //   Clear console on exit
    //   Reset info dialogs!
    // Hardware Options
    //   Use linear framebuffer
    //   Use VESA Palette setting
    //   Use virtual double-buffering
    // Universe
    //   Delete far planets
    //   Starcharts geometry...
    { N_("Backups"),                             0, "", NoEditor,         "" },
    {   N_("Result file backups..."),            1, "", BackupEditor,     "Backup.Result,%d/backups/result%p.%t" },
    {   N_("UTIL.DAT file backups..."),          1, "", BackupEditor,     "Backup.Util,%d/backups/util%p.%t" },
    {   N_("Turn file backups..."),              1, "", BackupEditor,     "Backup.Turn,%d/backups/turn%p.%t" },
    {   N_("Starchart database backups..."),     1, "", BackupEditor,     "Backup.Chart,%d/backups/chart%p.%t" },
    {   N_("Scripts/Auto Task backups..."),      1, "", BackupEditor,     "Backup.Script,%d/backups/script%p.%t" },
};

static void
addOptionByName(WConfigStorageControl& ctl, Config& cfg, string_t name)
{
    if (ConfigOption* opt = cfg.getConfigOption(name)) {
        ctl.addOption(*opt);
    }
}



WConfigBackupEditor::WConfigBackupEditor(Config& cfg)
    : cfg(cfg)
{ }

string_t
WConfigBackupEditor::toString(string_t arg)
{
    string_t name, deflt;
    strSplit(arg, name, deflt, ",");
    ConfigStringOption* opt = dynamic_cast<ConfigStringOption*>(cfg.getConfigOption(name));
    if (opt == 0) {
        return string_t();
    } else {
        string_t value = (*opt)();
        if (value.empty()) {
            return _("disabled (no backup)");
        } else if (value == deflt) {
            return _("enabled (standard file name)");
        } else {
            return format(_("custom: %s"), value);
        }
    }
}

void
WConfigBackupEditor::edit(string_t arg)
{
    string_t name, deflt;
    strSplit(arg, name, deflt, ",");
    ConfigStringOption* opt = dynamic_cast<ConfigStringOption*>(cfg.getConfigOption(name));
    if (opt) {
        WConfigBackupEditorWindow win(*opt, deflt);
        ui_root->centerWidget(win);
        ui_root->runModal(win);
    }
}


WConfigBackupEditorWindow::WConfigBackupEditorWindow(ConfigStringOption& opt, string_t def)
    : UIWindow(_("Backup Setting"), BLUE_WINDOW, COLOR_GRAY),
      opt(opt),
      def(def),
      h(),
      selection(0),
      input(0)
{
    init();
}

void
WConfigBackupEditorWindow::init()
{
    // VBox
    //   UIRadioButton "disabled"
    //   UIRadioButton "enabled"
    //   UIRadioButton "custom"
    //   HBox
    //     spacer
    //     UIInputLine
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer

    // Input line
    input = &h.add(new UIInputLine(1000, 20, 0, UIInputLine::ilf_Framed, 2003));
    input->setText(opt());

    // Pick default value
    if (opt().empty()) {
        selection.set(0);
        input->setText(def);
    } else if (opt() == def) {
        selection.set(1);
    } else {
        selection.set(2);
    }

    // Build window
    setLayout(UIVBoxLayout::instance5);
    add(h.add(new UIRadioButton(2000, _("disabled (no backup)"), 'd', selection, 0)));
    add(h.add(new UIRadioButton(2001, _("enabled (standard file name)"),  'e', selection, 1)));
    add(h.add(new UIRadioButton(2002, _("custom file name"),   'c', selection, 2)));

    UIGroup& group4 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group4.add(h.add(new UIStaticText("     ", UIStaticText::StaticTitle)));
    group4.add(*input);
    add(group4);

    UIGroup& group5 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group5.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &WConfigBackupEditorWindow::onOK));
    group5.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    group5.add(h.add(new UISpacer()));
    group5.add(h.add(new UIButton(_("Help"), 'h')));
    add(group5);

    add(h.add(new UIFocusIterator(2000, 2003, UIFocusIterator::fi_Tab | UIFocusIterator::fi_Vert)));

    selection.sig_changed.add(this, &WConfigBackupEditorWindow::onMove);
    getWidgetById(2000 + selection.get())->setFocusedWidget();
    add(h.add(new WHelpWidget("pcc2:settings:backup")));

    pack();
    onMove();
}

void
WConfigBackupEditorWindow::onMove()
{
    setStateOf(2003, st_Disabled, selection.get() != 2);
}

void
WConfigBackupEditorWindow::onOK()
{
    switch (selection.get()) {
     case 0:
        opt.setAndMarkUpdated(string_t(), ConfigOption::srcUser);
        break;
     case 1:
        opt.setAndMarkUpdated(def, ConfigOption::srcUser);
        break;
     case 2:
        opt.setAndMarkUpdated(input->getText(), ConfigOption::srcUser);
        break;
    }
    stop(1);
}

string_t
WConfigStarchartEditor::toString(string_t /*arg*/)
{
    return _("(dialog)");
}

void
WConfigStarchartEditor::edit(string_t /*arg*/)
{
    doChartConfig();
}


WConfigBooleanEditor::WConfigBooleanEditor(Config& cfg)
    : cfg(cfg)
{ }

string_t
WConfigBooleanEditor::toString(string_t arg)
{
    if (ConfigOption* opt = cfg.getConfigOption(arg)) {
        return opt->toString();
    } else {
        return string_t();
    }
}

void
WConfigBooleanEditor::edit(string_t arg)
{
    ConfigOption* opt = cfg.getConfigOption(arg);
    if (ConfigIntOption* io = dynamic_cast<ConfigIntOption*>(opt)) {
        io->set(!(*io)());
        io->markUpdated(ConfigOption::srcUser);
    } else if (ConfigIntRefOption* iro = dynamic_cast<ConfigIntRefOption*>(opt)) {
        iro->set(!(*iro)());
        iro->markUpdated(ConfigOption::srcUser);
    }
}


string_t
WConfigLabelEditor::toString(string_t /*arg*/)
{
    return _("(dialog)");
}

void
WConfigLabelEditor::edit(string_t /*arg*/)
{
    editLabelExpressions();
}





WPreferencesDialog::WPreferencesDialog(Config& cfg)
    : UIWindow(_("Settings"), BLUE_WINDOW, COLOR_GRAY),
      cfg(cfg),
      settings(0),
      details(0),
      editor(0),
      storage(0),
      editors()
{
    init();
}

void
WPreferencesDialog::init()
{
    // VBox
    //   UICardGroupTab
    //   HBox
    //     UICardGroup
    //       UITreeWidget "Options"
    //       WConfigView "Details"
    //     VBox
    //       WConfigEditor
    //       UISpacer            / later: doc
    //       WConfigStorageControl
    //   HBox
    //     UIButton "OK"
    //     UISpacer
    //     UIButton "Help"

    // Set up editors
    editors.push_back(0);
    editors.push_back(new WConfigOptionEditor(cfg));
    editors.push_back(new WConfigBackupEditor(cfg));
    editors.push_back(new WConfigStarchartEditor());
    editors.push_back(new WConfigBooleanEditor(cfg));
    editors.push_back(new WConfigLabelEditor());

    // Create these first so they die last, and card_group's sig_focus_changed
    // does not get in our way during destruction.
    settings = &h.add(new UITreeWidget(1000, 20, 25*font_ems[FONT_NORMAL]));
    details = &h.add(new WConfigView(1001, 20, 25*font_ems[FONT_NORMAL]));

    UICardGroup&    card_group = h.add(new UICardGroup());
    UICardGroupTab& card_tabs  = h.add(new UICardGroupTab(card_group));

    UIGroup& group2  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group22 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group3  = h.add(new UIGroup(UIHBoxLayout::instance5));

    setLayout(UIVBoxLayout::instance5);
    card_tabs.addTab(1000, UIKeyString(_("Options")));
    card_tabs.addTab(1001, UIKeyString(_("Configuration File")));
    add(card_tabs);
    for (uint_t i = 0; i < countof(config_tree); ++i) {
        settings->addItem(i, config_tree[i].level, _(config_tree[i].name));
    }
    settings->updateAfterModification();
    settings->sig_moved.add(this, &WPreferencesDialog::onMove);
    card_group.add(*settings);
    details->setNameColumnWidth(13);
    details->setContent(cfg);
    details->setHighlightSource(ConfigOption::srcUser);
    details->sig_moved.add(this, &WPreferencesDialog::onMove);
    card_group.add(*details);
    card_group.sig_focus_changed.add(this, &WPreferencesDialog::onMove);
    group2.add(UIFrameGroup::wrap(h, card_group));

    editor = &h.add(new WConfigEditor());
    group22.add(*editor);
    group22.add(h.add(new UISpacer()));
    storage = &h.add(new WConfigStorageControl(true));
    group22.add(*storage);
    group2.add(group22);
    add(group2);

    group3.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());
    group3.add(h.add(new UISpacer()));
    group3.add(h.add(new UIButton(_("Help"), 'h')));
    add(group3);
    add(h.add(new WHelpWidget("pcc2:settings")));

    pack();

    details->observe(editor->sig_change);
    details->observe(storage->sig_change);
    settings->setFocusedWidget();
    editor->sig_change.add(this, &WPreferencesDialog::onMove);

    onMove();
}

void
WPreferencesDialog::onMove()
{
    if (settings->hasState(st_Focused)) {
        storage->clear();

        const ConfigItem& item = config_tree[settings->getItemId(settings->getCurrentItemIndex())];
        editor->set(_(item.name), editors[item.editor], item.arg);
        switch (item.editor) {
         case NoEditor:
            break;
         case DefaultEditor:
         case BoolEditor:
            addOptionByName(*storage, cfg, item.arg);
            break;
         case BackupEditor:
            addOptionByName(*storage, cfg, strFirst(item.arg, ","));
            break;
         case StarchartEditor:
            for (int i = 0, n = getNumChartOptions(); i < n; ++i) {
                storage->addOption(getChartOption(i));
            }
            for (int i = 0; i < ConfigMarkerOption::NUM_CANNED_MARKERS; ++i) {
                storage->addOption(ConfigMarkerOption::getCannedMarker(i));
            }
            for (int i = 0; i < lm_MAX; ++i) {
                storage->addOption(getLockModeConfigOption(TLockMode(i)));
            }
            break;
         case LabelEditor:
            addOptionByName(*storage, cfg, "Label.Ship");
            addOptionByName(*storage, cfg, "Label.Planet");
            break;
        }
    } else if (details->hasState(st_Focused)) {
        storage->clear();
        if (ConfigOption* opt = details->getCurrentOption()) {
            storage->addOption(*opt);
            editor->set(opt->getName(), editors[DefaultEditor], opt->getName());
        } else {
            editor->set(string_t(), 0, string_t());
        }
    }
}

/** Display the preferences dialog. */
void
doPreferencesDialog()
{
    WPreferencesDialog dlg(getUserPreferences());
    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
}
