/**
  *  \file client/dialogs/simfunction.cc
  *  \brief Simulator Ship Function Editor
  */

#include <cctype>
#include "client/dialogs/simfunction.h"
#include "client/widgets/helpwidget.h"
#include "client/wtile.h"
#include "game/sim.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/strtab.h"
#include "ui/util.h"

namespace {
    const int NUM_FUNCTIONS = 7;

    class WSimFunctionEditor;

    class WSimFunctionTile : public WTile {
        WSimFunctionEditor& editor;
        void init();
        void computeLayout();
        ObjectHolder h;
        int left_x, left_wi;
        int right_x;
     public:
        WSimFunctionTile(WSimFunctionEditor& editor);
        void getLayoutInfo(LayoutInfo& info);
        void drawData(GfxCanvas& can);
    };

    class WSimFunctionEditor : public UIWindow {
        GSimShip& ship;
        ObjectHolder holder;
        int new_flags;
        WTile* ptile;

        void toggle(GSimFunction sf);
        void init();
     public:
        WSimFunctionEditor(GSimShip& sh);
        bool handleEvent(const UIEvent& e, bool second_pass);
        const char* getFunctionType(GSimFunction sf) const;
        int getNewFlags() const
            { return new_flags; }
    };

    class WSimFunctionButton : public UIButton {
     public:
        WSimFunctionButton(const string_t name, UIKey key)
            : UIButton(name, key)
            { setState(bf_SmallSize, true); }
    };
}

static const char KEYS[NUM_FUNCTIONS] = { 'I', 'F', 'C', 'K', 'B', 'T', 'E' };

static const char* const LABELS[NUM_FUNCTIONS] = {
    N_("Planet Immunity"),
    N_("Full Weaponry"),
    N_("Commander"),
    N_("3\xC3\x97 Beam Kill"),
    N_("2\xC3\x97 Beam Charge"),
    N_("2\xC3\x97 Torp Charge"),
    N_("Elusive"),
};

static const char* const VALUES[] = {
    N_("no"),
    N_("yes"),
    N_("default (=no)"),
    N_("default (=yes)"),
};


WSimFunctionTile::WSimFunctionTile(WSimFunctionEditor& editor)
    : WTile(GfxPoint(1, 1)),
      editor(editor)
{
    init();
}

void
WSimFunctionTile::init()
{
    using std::tolower;
    const int ly = font_heights[FONT_TITLE];
    setLayout(LeftJustified, 0, false);
    for (int i = 0; i < NUM_FUNCTIONS; ++i) {
        addControl(GfxRect(0, ly*i+1, ly-2, ly-2),
                   h.add(new UIButton(string_t(&KEYS[i], 1),
                                      tolower(KEYS[i]))));
    }
}

void
WSimFunctionTile::computeLayout()
{
    int right_wi;

    left_wi = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(LABELS, countof(LABELS))) + 20;
    right_wi = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(VALUES, countof(VALUES)));

    left_x = font_heights[FONT_TITLE] + 5;
    right_x = left_x + left_wi + right_wi + 5;
}

void
WSimFunctionTile::getLayoutInfo(LayoutInfo& info)
{
    computeLayout();
    info.min_size = info.pref_size =
        GfxPoint(right_x + 5, NUM_FUNCTIONS * font_heights[FONT_TITLE]);
    info.flags = 0;
}

void
WSimFunctionTile::drawData(GfxCanvas& can)
{
    computeLayout();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    int right_wi = right_x - left_x - left_wi;
    for (int i = 0; i < NUM_FUNCTIONS; ++i) {
        // Left side
        ctx.setTransparentBackground();
        ctx.setTextAlign(0, 1);
        ctx.setColor(UIColor::tc_Static);
        int y = (font_heights[FONT_TITLE] * (2*i+1)) / 2;
        outText(ctx, left_x, y, string_t(_(LABELS[i])) + ":");

        // Right side
        ctx.setSolidBackground();
        ctx.setTextAlign(2, 1);
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, right_x, y, right_wi, editor.getFunctionType(GSimFunction(i)));
    }
}

WSimFunctionEditor::WSimFunctionEditor(GSimShip& ship)
    : UIWindow(_("Ship Abilities"), BLUE_WINDOW, COLOR_GRAY),
      ship(ship),
      new_flags(ship.getFlags()),
      ptile(0)
{
    init();
}

void
WSimFunctionEditor::init()
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));

    WTile& tile = holder.add(new WSimFunctionTile(*this));

    buttons.add(holder.add(new UIButton(cm_Enter,  _("OK"), SDLK_RETURN)).addStop());
    buttons.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    buttons.add(holder.add(new UISpacer()));
    buttons.add(holder.add(new UIButton(_("H"), 'h')));

    add(tile);
    add(buttons);
    add(holder.add(new WHelpWidget("pcc2:simfunctions")));
    pack();

    // WSimPlanetEditor uses a signal for this; since we have just one client, we can
    // as well just keep a pointer.
    ptile = &tile;
}

void
WSimFunctionEditor::toggle(GSimFunction sf)
{
    const int32_t value = GSimShip::simf_to_value_bit[sf];
    const int32_t set = GSimShip::simf_to_set_bit[sf];
    const int32_t flags = new_flags;
    if ((flags & set) == 0) {
        // default => Yes
        new_flags |= (value | set);
    } else if ((flags & value) != 0) {
        // Yes => No
        new_flags &= ~value;
    } else {
        // No => Default
        new_flags &= ~set;
    }

    if (ptile)
        ptile->drawWidget();
}

bool
WSimFunctionEditor::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        for (int i = 0; i < NUM_FUNCTIONS; ++i) {
            if (e.key.code == uint32_t(std::tolower(KEYS[i]))) {
                toggle(GSimFunction(i));
                return true;
            }
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

const char*
WSimFunctionEditor::getFunctionType(GSimFunction sf) const
{
    const int32_t value = GSimShip::simf_to_value_bit[sf];
    const int32_t set = GSimShip::simf_to_set_bit[sf];
    const int32_t flags = new_flags;

    int index = 0;
    if ((flags & set) != 0) {
        if ((flags & value) != 0) {
            index = 1;
        } else {
            index = 0;
        }
    } else {
        if (ship.hasImpliedHullFunction(sf)) {
            index = 3;
        } else {
            index = 2;
        }
    }
    return _(VALUES[index]);
}

/** Edit special abilities of the specified simulator ship.
    \param sh Ship to edit
    \retval true Dialog was confirmed, ship possibly changed
    \retval false Dialog was cancelled, ship unchanged */
bool
editSimulatorShipFunctions(GSimShip& sh)
{
    WSimFunctionEditor fedit(sh);
    ui_root->centerWidget(fedit);
    ui_root->add(fedit);
    int result = ui_root->run();
    ui_root->close(fedit);
    if (result == cm_Enter) {
        sh.setFlags(fedit.getNewFlags());
        return true;
    } else {
        return false;
    }
}

string_t
describeFunctions(const GSimShip& sh)
{
    string_t result;
    for (int i = 0; i < NUM_FUNCTIONS; ++i) {
        if (sh.hasShipFunction(GSimFunction(i))) {
            if (!result.empty()) {
                result += ", ";
            }
            result += _(LABELS[i]);
        }
    }
    if (result.empty()) {
        result = _("none");
    }
    return result;
}
