/**
  *  \file client/scr-score.cc
  *  \brief Score Display Screen
  *
  *  Handling of aggregation modes: WScoreTable can display the
  *  aggregated modes, but this doesn't have much sense. We therefore
  *  maintain only the main, non-aggregated mode in the WScoreTable,
  *  and an additional flag whether the WScoreChart's should use
  *  aggregation.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <algorithm>
#include "game/stat.h"
#include "client/wscore.h"
#include "util/nls.h"
#include "ui/window.h"
#include "ui/msgbox.h"
#include "ui/layout.h"
#include "gfx/palette.h"
#include "ui/iconbox.h"
#include "game/host.h"
#include "cpluslib/format.h"
#include "ui/listbox.h"
#include "client/widgets/racelist.h"
#include "game/team.h"
#include "ui/util.h"
#include "ui/widgets/statictext.h"
#include "util/recode.h"
#include "ui/widgets/button.h"
#include "ui/widgets/cardgroup.h"
#include "ui/widgets/cardtab.h"
#include "client/widgets/helpwidget.h"
#include "io/io.h"

namespace {
    enum {
        cm_SetMode = 1000
    };

    /// Kind of score page.
    enum ScoreKind {
        Table,                  ///< WScoreTable.
        Chart                   ///< WScoreChart.
    };

    /// Description of a score page.
    struct ScoreTab {
        ScoreKind kind;         ///< Kind of page.
        string_t  name;         ///< Name of page. Shown on button and in heading.
        uint8     key;          ///< Hotkey. Single ASCII letter.
        WScore    score;        ///< Score. Describes the score on a WScoreChart.

        ScoreTab(ScoreKind kind, string_t name, WScore score);
        ScoreTab(WScore score);
        ~ScoreTab();
    };

    /** Description of all possible selections on the score screen. */
    typedef std::vector<ScoreTab> ScoreTabVector;

    /** Score screen page selector. Offers a list of score pages. Also adds keyboard
        handling. */
    class WScoreIconBox : public UIIconBox {
        ScoreTabVector& tabs;
        string_t getSubTitle(int number);
     public:
        WScoreIconBox(ScoreTabVector& tabs);
        int getItemWidth(int number);
        int getNumberOfItems();
        void drawItem(int number, GfxCanvas& can, const GfxRect& r, ItemState state);
        bool handleEvent(const UIEvent& event, bool second_pass);
    };

    /** Score window. Displays contents of a GStatFile in a user-definable selection
        of views (pages). */
    class WScoreWindow : public UIWindow {
        ObjectHolder h;
        WScoreTable& score_table;
        WScoreChart& score_chart;
        WScoreIconBox& iconbox;
        WRaceList& race_list;
        const ScoreTabVector& tabs;
        UICardGroup& cardgroup;
        int current_page;
        UIStaticText title;
        bool cumulative;

        void init(const GStatFile& file);
     public:
        WScoreWindow(const GStatFile& file, ScoreTabVector& tabs);

        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        bool handleEvent(const UIEvent& event, bool second_pass);

        void openPage(int page, int player);
        void changeToPage(int page);

        void setAggregation(WScoreAggregation aggreg);
    };
}

/** General constructor. */
ScoreTab::ScoreTab(ScoreKind kind, string_t name, WScore score)
    : kind(kind), name(name), key(0), score(score)
{ }

/** Constructs a Chart descriptor. Name and score are taken from \c score. */
ScoreTab::ScoreTab(WScore score)
    : kind(Chart), name(score.getName()), key(0), score(score)
{ }

/** Destructor. */
ScoreTab::~ScoreTab()
{ }

/***************************** WScoreIconBox *****************************/

/** Constructor. */
WScoreIconBox::WScoreIconBox(ScoreTabVector& tabs)
    : UIIconBox(GfxPoint(400, font_heights[FONT_NORMAL] + font_heights[FONT_TITLE] + 10)),
      tabs(tabs)
{ }

/** Get subtitle. This is displayed below the page name, and includes type
    and hotkey for the page. */
string_t
WScoreIconBox::getSubTitle(int number)
{
    string_t subtitle = tabs[number].kind == Table ? _("Table") : _("Graph");
    if (tabs[number].key != 0) {
        subtitle += " [";
        subtitle += tabs[number].key;
        subtitle += "]";
    }
    return subtitle;
}

int
WScoreIconBox::getItemWidth(int number)
{
    return std::max(fonts[FONT_TITLE]->getTextWidth(tabs[number].name),
                    fonts[FONT_NORMAL]->getTextWidth(getSubTitle(number))) + 10;
}

int
WScoreIconBox::getNumberOfItems()
{
    return tabs.size();
}

void
WScoreIconBox::drawItem(int number, GfxCanvas& can, const GfxRect& r, ItemState state)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    if (state == Normal) {
        drawBackground(ctx, r);
        ctx.setColor(UIColor::tc_Static);
        ctx.useFont(*fonts[FONT_TITLE]);
        outText(ctx, r.x + 5, r.y + 5, tabs[number].name);
        ctx.useFont(*fonts[FONT_NORMAL]);
        outText(ctx, r.x + 5, r.y + 5 + font_heights[FONT_TITLE], getSubTitle(number));
    } else {
        drawBackground(ctx, GfxRect(r.x,           r.y,           1, 1));
        drawBackground(ctx, GfxRect(r.x + r.w - 1, r.y,           1, 1));
        drawBackground(ctx, GfxRect(r.x,           r.y + r.h - 1, 1, 1));
        drawBackground(ctx, GfxRect(r.x + r.w - 1, r.y + r.h - 1, 1, 1));

        ctx.setRawColor(state == Hover ? standard_colors[COLOR_GRAYSCALE+6] : standard_colors[COLOR_BLACK]);
        drawBar(ctx, GfxRect(r.x,   r.y+1,     r.w,   r.h-2));
        drawBar(ctx, GfxRect(r.x+1, r.y,       r.w-2, 1));
        drawBar(ctx, GfxRect(r.x+1, r.y+r.h-1, r.w-2, 1));

        ctx.setRawColor(state == Hover ? standard_colors[COLOR_BLACK] : standard_colors[COLOR_WHITE]);
        ctx.useFont(*fonts[FONT_TITLE]);
        outText(ctx, r.x + 5, r.y + 5, tabs[number].name);
        ctx.useFont(*fonts[FONT_NORMAL]);
        outText(ctx, r.x + 5, r.y + 5 + font_heights[FONT_TITLE], getSubTitle(number));
    }
}

bool
WScoreIconBox::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        if (event.key.code == SDLK_TAB || event.key.code == SDLK_RIGHT) {
            // Tab: cycle forward
            setSelectedWidget();
            int n = getCurrent()+1;
            if (n >= getNumberOfItems())
                setCurrent(0);
            else
                setCurrent(n);
            return true;
        } else if (event.key.code == SDLK_TAB + ss_Shift || event.key.code == SDLK_LEFT) {
            // Shift-Tab: cycle backward
            setSelectedWidget();
            int n = getCurrent()-1;
            if (n < 0)
                setCurrent(getNumberOfItems()-1);
            else
                setCurrent(n);
            return true;
        } else {
            // Hot-keys to select individual pages
            for (ScoreTabVector::size_type i = 0; i < tabs.size(); ++i) {
                if (tabs[i].key != 0 && event.key.code == tabs[i].key) {
                    setSelectedWidget();
                    setCurrent(i);
                    return true;
                }
            }
        }
    }
    return UIIconBox::handleEvent(event, second_pass);
}

/****************************** WScoreWindow *****************************/

/** Score Window Constructor.
    \param file File to display
    \param tabs View selection. Describes all possible views (pages). */
WScoreWindow::WScoreWindow(const GStatFile& file, ScoreTabVector& tabs)
    : UIWindow(_("Scores"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      score_table(h.add(new WScoreTable(file))),
      score_chart(h.add(new WScoreChart(file, 0, file.getNumTurns()))),
      iconbox(h.add(new WScoreIconBox(tabs))),
      race_list(h.add(new WRaceList(400, WRaceList::FlowLayout | WRaceList::ShowRaceNames, 1, NUM_PLAYERS))),
      tabs(tabs),
      cardgroup(h.add(new UICardGroup())),
      current_page(0),
      title("", UIStaticText::StaticTitle, 1),
      cumulative(false)
{
    init(file);
}

void
WScoreWindow::init(const GStatFile& file)
{
    /* Widget structure:
         Window [UIVBoxLayout]
           UICardGroup
             UIGroup [UIVBoxLayout]
               WScoreTable
               UIGroup [UIHBoxLayout]
                 UIButton "+"
                 UIButton "-"
                 UISpacer
             UIGroup [UIVBoxLayout]
               WScoreChart
               WRaceList
           WScoreIconBox
           UIGroup [UIHBoxLayout]
             UISpacer
             UIButton "Close" */

    setLayout(UIVBoxLayout::instance5);

    score_table.addDefaultColumns();

    /* create widgets */
    UIGroup& table_page         = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& table_button_group = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& chart_page         = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& button_group       = h.add(new UIGroup(UIHBoxLayout::instance5));

    UIButton& table_next   = h.add(new UIButton(0, "+", '+'));
    UIButton& table_prev   = h.add(new UIButton(0, "-", '-'));
    UIButton& close_button = h.add(new UIButton(0, _("Close"), SDLK_ESCAPE));

    /* connect widgets */
    cardgroup.add(table_page);
    cardgroup.add(chart_page);

    table_page.add(score_table);
    table_page.add(table_button_group);

    table_button_group.add(table_next);
    table_button_group.add(table_prev);
    table_button_group.add(h.add(new WScoreTableTimeDisplay(score_table, file)));
    table_button_group.add(h.add(new UISpacer()));
    table_button_group.add(h.add(new UIButton(cm_SetMode, "D", 'd')).addCommand());
    table_button_group.add(h.add(new WScoreTableConfigDisplay(score_table)));

    chart_page.add(score_chart);
    chart_page.add(race_list);

    button_group.add(close_button);
    button_group.add(h.add(new UISpacer()));
    button_group.add(h.add(new UIButton(_("Help"), 'h')));

    add(title);
    add(cardgroup);
    add(iconbox);
    add(button_group);
    add(h.add(new WHelpWidget("pcc2:scores")));

    /* connect events */
    table_prev.addCall(&score_table, &WScoreTable::previousTurn);
    table_next.addCall(&score_table, &WScoreTable::nextTurn);
    close_button.addStop();

    iconbox.sig_changed.add(this, &WScoreWindow::changeToPage);
    race_list.sig_click.add(&score_chart, &WScoreChart::setCurrentPlayer);

    pack();
}

bool
WScoreWindow::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_SetMode) {
        /* Build menu */
        UIStringList list;
        for (int i = 0; i < score_table.NUM_MODES; ++i)
            list.add(i, score_table.getModeName(WScoreTable::Mode(i), score_table.getCurrentRecord()));
        if (isTeamConfigured()) {
            list.add(1000, _("Scores by player"));
            list.add(1001, _("Scores by team"));
        }

        UIStandardListbox listbox(0, 1, getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
        listbox.scrollToKey(score_table.getMode());
        if (listbox.doStandardMenu(sender)) {
            if (listbox.getKey() == 1000) {
                setAggregation(ScoreByPlayer);
            } else if (listbox.getKey() == 1001) {
                setAggregation(ScoreByTeam);
            } else {
                score_table.setMode(WScoreTable::Mode(listbox.getKey()));
            }
        }
        return true;
    }
    return UIWindow::handleCommand(cmd, sender);
}

bool
WScoreWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == 'y') {
        setSelectedWidget();
        if (score_table.getAggregation() == ScoreByPlayer && isTeamConfigured())
            setAggregation(ScoreByTeam);
        else
            setAggregation(ScoreByPlayer);
        return true;
    }
    if (event.type == ev_Key && event.key.code == 'x') {
        /* Toggle cumulative mode, only when chart is active */
        if (score_chart.hasState(st_Focused)) {
            setSelectedWidget();
            cumulative = !cumulative;
            setAggregation(score_table.getAggregation());
        }
        return true;
    }
    return UIWindow::handleEvent(event, second_pass);
}

/** Open a page.
    \param page Page number (0-based index into tabs).
    \param player Player to highlight. Can be 0. */
void
WScoreWindow::openPage(int page, int player)
{
    const ScoreTab& the_page = tabs[page];

    if (the_page.kind == Table) {
        // It's the table
        score_table.setCurrentPlayer(player);
        score_table.setFocusedWidget();
    } else {
        // It's a chart
        score_chart.setScore(the_page.score);
        score_chart.setCurrentPlayer(player);
        score_chart.setFocusedWidget();
    }
    iconbox.setCurrent(page);
    title.setText(the_page.name);

    // We have to remember the current page, because iconbox's current already
    // has changed when we receive the changeToPage event.
    current_page = page;
}

/** Change to a page. Like openPage(), but preserves the player number from
    the previously-shown page.
    \param page Page number to switch to. */
void
WScoreWindow::changeToPage(int page)
{
    // first, read out current page's player
    int current_player = tabs[current_page].kind == Table ? score_table.getCurrentPlayer() : score_chart.getCurrentPlayer();

    // switch to page
    openPage(page, current_player);
}

/** Set aggregation mode. */
void
WScoreWindow::setAggregation(WScoreAggregation aggreg)
{
    score_table.setAggregation(aggreg);
    score_chart.setAggregation(cumulative ? makeCumulative(aggreg) : aggreg);

    if (aggreg == ScoreByPlayer)
        race_list.setFlags(WRaceList::FlowLayout | WRaceList::ShowRaceNames);
    else
        race_list.setFlags(WRaceList::FlowLayout | WRaceList::ShowTeamNames);
}

/********************************** Main *********************************/

/** Generate view selection for a score file.
    \param file [in] Score file
    \param tabs [out] Views

    \todo If the score file contains a real score (ScoreId_Score), display that
    as "Score", and list "Tim-Score" (politically-correct name to be defined) later.
    Should probably also show the name of the score. */
static void
generateScoreTabs(const GStatFile& file, ScoreTabVector& tabs)
{
    using namespace std;        // for ctype

    /* add Overview list */
    tabs.push_back(ScoreTab(Table, _("Overview"), WScore("")));

    if (file.getNumTurns() > 1) {
        /* add regular scores */
        GStatIndex pi = file.getIndexForScore(ScoreId_Planets);
        GStatIndex bi = file.getIndexForScore(ScoreId_Bases);
        GStatIndex fi = file.getIndexForScore(ScoreId_Freighters);
        GStatIndex ci = file.getIndexForScore(ScoreId_Capital);
        GStatIndex qi = file.getIndexForScore(ScoreId_BuildPoints);

        /* Order in PCC 1.x is: Scores, Capital Ships, Freighters, Total Ships,
           Starbases, Planets, PAL/PBP */
        if (pi != file.nil && bi != file.nil && fi != file.nil && ci != file.nil)
            tabs.push_back(ScoreTab(WScore(_("Score")).add(pi, 10).add(bi, 120).add(fi, 1).add(ci, 10)));
        if (pi != file.nil)
            tabs.push_back(ScoreTab(WScore(_("Planets")).add(pi, 1)));
        if (fi != file.nil)
            tabs.push_back(ScoreTab(WScore(_("Freighters")).add(fi, 1)));
        if (ci != file.nil)
            tabs.push_back(ScoreTab(WScore(_("Capital Ships")).add(ci, 1)));
        if (ci != file.nil && fi != file.nil)
            tabs.push_back(ScoreTab(WScore(_("Total Ships")).add(ci, 1).add(fi, 1)));
        if (bi != file.nil)
            tabs.push_back(ScoreTab(WScore(_("Bases")).add(bi, 1)));
        if (qi != file.nil) {
            if (host.isPBPGame())
                tabs.push_back(WScore(_("PBPs")).add(qi, 1));
            else
                tabs.push_back(WScore(_("PAL")).add(qi, 1));
        }

        /* add remaining scores */
        for (GStatIndex i = 0; i < file.getNumScores(); ++i) {
            int16 id = file.getScoreId(i);
            if (id != ScoreId_Planets && id != ScoreId_Bases && id != ScoreId_Freighters
                && id != ScoreId_Capital && id != ScoreId_BuildPoints)
            {
                /* it's a column we didn't yet handle. Figure out its name */
                string_t name;
                if (id == ScoreId_MinesAllowed)
                    name = _("Minefields Allowed");
                else if (id == ScoreId_MinesLaid)
                    name = _("Minefields Laid");
                else if (const TScoreDescription* desc = file.getDescriptionForScoreId(id))
                    name = convertGameToUtf8(getBasicString(desc->name));
                else
                    name = format(_("Score #%d"), id);

                tabs.push_back(WScore(name).add(i, 1));
            }
        }
    }

    /* assign keys */
    const uint32 used_letters_externally
        = (1 << ('d'-'a'))                      // display mode menu
        | (1 << ('y'-'a'))                      // team toggle
        | (1 << ('x'-'a'))                      // cumulative toggle
        | (1 << ('h'-'a'));                     // help

    uint32 used_letters = used_letters_externally;
    for (ScoreTabVector::iterator i = tabs.begin(); i != tabs.end(); ++i) {
        /* First check whether any word starter can be a key */
        char c = 0;
        bool look_at_this_char = true;
        for (string_t::size_type si = 0; si < i->name.size(); ++si) {
            char this_char = tolower(static_cast<unsigned char>(i->name[si]));
            if (look_at_this_char && this_char >= 'a' && this_char <= 'z'
                && (used_letters & (1 << (this_char-'a'))) == 0)
            {
                /* ok */
                c = this_char;
                break;
            }

            look_at_this_char = (this_char == ' ');
        }

        /* If we didn't find one, take any free one */
        if (c == 0) {
            for (char ci = 'a'; ci <= 'z'; ++ci) {
                if ((used_letters & (1 << (ci-'a'))) == 0) {
                    c = ci;
                    break;
                }
            }
        }

        /* Still didn't find one? This means letters used up. Start anew. */
        if (c == 0) {
            used_letters = used_letters_externally;
            c = 'a';
        }

        /* Found one, assign it */
        i->key = c;
        used_letters |= 1 << (c - 'a');
    }
}

/** Display score screen.
    \param file   Statistics file to display
    \param player Player to highlight */
void
doScoreScreen(const GStatFile& file, int player)
{
    /* Generate layout */
    ScoreTabVector pages;
    generateScoreTabs(file, pages);

    /* Generate widgets */
    WScoreWindow window(file, pages);
    window.openPage(0, player);

    /* Run */
    ui_root->centerWidget(window);
    ui_root->add(window);
    ui_root->run();
    ui_root->close(window);
}
