/**
  *  \file client/widgets/dirchoose.cc
  *  \brief Game Directory Chooser Widget
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/dirchoose.h"
#include "cpluslib/format.h"
#include "gfx/font.h"
#include "io/io.h"
#include "ui/msgbox.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/recode.h"
#include "resmgr/resmgr.h"

static int
getLevel(Ptr<Directory> dir)
{
    int level = 0;
    while (dir != 0 && level < 1000) {
        ++level;
        dir = dir->getParent();
    }
    return level;
}

/************************ WDirectoryChooser::Entry ***********************/

WDirectoryChooser::Entry::Entry(Action a, const string_t& name, Ptr<Directory> entry)
    : a(a), name(name), entry(entry)
{ }

WDirectoryChooser::Entry::~Entry()
{ }

/*************************** WDirectoryChooser ***************************/

/** \class WDirectoryChooser
    \brief Game Directory Chooser

    Widget that displays a directory list and lets the user choose from it. */

/** Initialize. Widget starts by displaying nothing.
    \param id widget id. */
WDirectoryChooser::WDirectoryChooser(int id)
    : UIListbox(id, 14, 300, 0),
      file_icons(requestPixmap("files")),
      current(0)
{
    sig_dblclick.add(this, &WDirectoryChooser::enterCurrent);
}

WDirectoryChooser::~WDirectoryChooser()
{ }

void
WDirectoryChooser::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    for (int line = from; line < to; ++line) {
        GfxRect area;
        prepareEntry(ctx, HighContrastStyle, line, &area);

        const int index = getTop() + line;
        if (index < int(content.size())) {
            if (file_icons) {
                static const int8_t ys[] = { 16, 32, 48, 64 };
                file_icons->blit(ctx.getCanvas(), area.x, area.y, GfxRect(0, ys[content[index].a], 16, 16));
            }
            area.consumeX(20);
            outTextF(ctx, area, content[index].name);
        }
    }
}

/** Enter current directory. When something goes wrong, tries to recover
    (= stay in current directory). */
void
WDirectoryChooser::enterCurrent()
{
    if (getCurrent() < int(content.size())) {
        switch (content[getCurrent()].a) {
         case aRescanWithoutGamelist:
            setContent(scanNoGamelist, current);
            break;
         case aChdir:
         case aParent:
            setContent(scanGamelist, content[getCurrent()].entry);
            break;
         case aShowRoots:
            setContent(scanRoots, 0);
            break;
        }
    }
}

/** Enter parent directory. */
void
WDirectoryChooser::enterParent()
{
    if (current != 0) {
        if (Ptr<Directory> p = current->getParent()) {
            setContent(scanGamelist, p);
        }
    }
}

/** Enter a directory.
    \param dir Directory. May be null to show roots. */
void
WDirectoryChooser::enterDirectory(Ptr<Directory> dir, bool oneup)
{
    if (dir) {
        Ptr<Directory> parent;
        if (oneup && (parent = dir->getParent()) != 0) {
            setContent(scanGamelist, parent);
            showDirectory(dir);
        } else {
            setContent(scanGamelist, dir);
        }
    } else {
        setContent(scanRoots, 0);
    }
}

/** Show directory. Places the cursor on dir (or one of its parents)
    shown in the current display. */
void
WDirectoryChooser::showDirectory(Ptr<Directory> dir)
{
    // Skip if dir is null
    if (!dir) {
        setCurrent(0);
        return;
    }

    // Find both directories' levels
    int current_level = getLevel(current);
    int dir_level     = getLevel(dir);

    // Adjust so that dir is exactly beneath current
    ++current_level;
    while (dir_level > current_level) {
        dir = dir->getParent();
        --dir_level;
    }

    // Match?
    int n = 0;
    if (dir_level == current_level) {
        for (std::vector<Entry>::size_type i = 0; i != content.size(); ++i) {
            if (content[i].a == aChdir && content[i].entry->getName() == dir->getName()) {
                n = i;
                break;
            }
        }
    }
    setCurrent(n);
}

/** Get current subdirectory (=the directory the cursor is on).
    \return directory, or 0 if none (e.g. if cursor is on "[Parent]" entry */
Ptr<Directory>
WDirectoryChooser::getSubdirectory()
{
    if (getCurrent() < int(content.size()) && content[getCurrent()].a == aChdir)
        return content[getCurrent()].entry;
    return 0;
}

/** Set content. This will read the directory and update display.
    If anything goes wrong, it will display a message box and back up.
    \param arg how to read the directory
    \param dir directory to read (can be null if arg is scanRoots) */
void
WDirectoryChooser::setContent(ScanArg arg, Ptr<Directory> dir)
{
    // Try it
    try {
        scan(arg, dir);
    }
    catch (FileProblemException& e) {
        // It failed. Try to back up.
        bool fail = true;
        Ptr<Directory> parent;
        while (dir != 0 && fail && (parent = dir->getParent()) != 0) {
            if (current != 0 && parent->getName() == current->getName()) {
                // Found the directory that we started in. So just don't change anything.
                fail = false;
            } else {
                // Try to read that directory.
                try {
                    dir = parent;
                    scan(scanGamelist, dir);
                    fail = false;
                }
                catch (...) { }
            }
        }

        // Still failed? Try to revert to original.
        if (fail && current != 0) {
            fail = false;
        }

        // Still failed? Try to read roots. This ought not to fail.
        if (fail) {
            try {
                scan(scanRoots, 0);
                fail = false;
            }
            catch (...) { }
        }

        // Finally, inform the user what happened
        messageBox(format(_("Unable to display the requested directory: %s: %s"), e.getFileName(), e.what()), _("Error"));
    }
}

/** Initialize directory. This will read the directory, and update
    display on success. Throws an exception if anything goes wrong,
    in this case, the content will not be modified.
    \param arg how to read the directory
    \param dir directory to read (can be null if arg is scanRoots) */
void
WDirectoryChooser::scan(ScanArg arg, Ptr<Directory> dir)
{
    std::vector<Entry> newContent;
    if (arg == scanRoots) {
        /* scan roots (=devices) */
        Ptr<Enumeration<Ptr<Directory> > > roots = Directory::getRoots();
        while (roots->hasMoreElements()) {
            Ptr<Directory> dir = roots->getNextElement();
            newContent.push_back(Entry(aChdir, dir->getTitle(), dir));
        }
    } else {
        if (Ptr<Directory> p = dir->getParent()) {
            newContent.push_back(Entry(aParent, _("[Up one level]"), p));
        }

        if (arg == scanGamelist && readGamelist(newContent, *dir)) {
            /* ok */
        } else {
            /* no game list requested or none found */
            readDirectory(newContent, *dir);
        }

        newContent.push_back(Entry(aShowRoots, _("[Other Places]"), 0));
    }

    Ptr<Directory> previous = current;
    content.swap(newContent);
    current = dir;
    setSize(content.size());
    showDirectory(previous);
    adjustPosition();
    drawWidget(false);

    sig_dirchange.raise();
    sig_moved.raise(getId());
}

/** Scan directory content.
    \param newContent [out] Content
    \param dir [in] Directory to read
    \throw FileProblemException directory cannot be read */
void
WDirectoryChooser::readDirectory(std::vector<Entry>& newContent, Directory& dir)
{
    const std::vector<Entry>::size_type first_index = newContent.size();
    Ptr<Enumeration<Ptr<DirectoryEntry> > > list = dir.getEntries();
    while (list->hasMoreElements()) {
        Ptr<DirectoryEntry> entry = list->getNextElement();
        if (entry->getFileType() == DirectoryEntry::FT_DIRECTORY) {
            newContent.push_back(Entry(aChdir, entry->getBaseName(), entry->openDir()));
        }
    }
    std::sort(newContent.begin() + first_index, newContent.end(), isEntryLT);
}

/** Read gamestat.dat file.
    \param newContent [out] Content
    \param dir [in] Directory to read
    \retval true File could be read, newContent has been updated
    \retval false File could not be read, newContent has not been modified */
bool
WDirectoryChooser::readGamelist(std::vector<Entry>& newContent, Directory& dir)
{
    try {
        // Open file
        Ptr<Stream> s = dir.openFile("gamestat.dat", Stream::C_READ);

        // Read content
        char buffer[8][206];
        s->readT(buffer[0], sizeof buffer);

        // Check content
        for (int i = 0; i < 8; ++i) {
            newContent.push_back(Entry(aChdir, " " + convertGameToUtf8(getBasicStringN(&buffer[i][184], 20)),
                                       dir.openSubdir(format("vpwork%d", i+1))));
        }
        newContent.push_back(Entry(aRescanWithoutGamelist, _("[Directory Content]"), 0));
        return true;
    }
    catch (FileProblemException& e) {
        return false;
    }
}

/** Compare two entries. */
bool
WDirectoryChooser::isEntryLT(const Entry& l, const Entry& r)
{
    return strCaseCmp(l.name, r.name) < 0;
}
