/**
  *  \file client/widgets/fleetmemlist.cc
  *  \brief Fleet Member List
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/widgets/fleetmemlist.h"
#include "game/mission.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "util/unicodechars.h"
#include "cpluslib/format.h"
#include "game/univ.h"
#include "game/global.h"
#include "game/ship.h"
#include "client/marks.h"

WFleetMemberList::WFleetMemberList(int lines, int width)
    : UIListbox(0, lines, width, 0),
      current_fleet(0)
{ }

WFleetMemberList::~WFleetMemberList()
{ }

void
WFleetMemberList::setFleet(const int fid)
{
    // clear, and re-initialize only when non-null
    int newCurrent = 0;
    bool change = (fid != current_fleet);

    content.clear();
    current_fleet = fid;

    if (fid != 0) {
        // add leader
        content.push_back(fid | fLeader);
        GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        GShip& leader = univ.getShip(fid);

        // add members
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            GShip& mem = univ.getShip(i);
            if (fid != i && mem.isPlayable(GObject::Playable) && mem.getFleetNumber() == fid) {
                uint16_t value = i;
                if (mem.getPos() != leader.getPos()) {
                    value |= fAway;
                }
                if (i == current_ship_selection.getCurrentIndex()) {
                    newCurrent = content.size();
                }
                content.push_back(value);
            }
        }

        // process tow/towee
        for (std::vector<uint16_t>::size_type i = 0; i < content.size(); ++i) {
            GShip& mem = univ.getShip(content[i] & fIdMask);
            if (mem.getMission() == GMission::msn_Tow) {
                for (std::vector<uint16_t>::size_type j = 0; j < content.size(); ++j) {
                    if (j != i && (content[j] & fIdMask) == mem.getTowId()) {
                        content[i] |= fTower;
                        content[j] |= fTowed;
                        break;
                    }
                }
            }
        }
    }

    // configure list
    setSize(content.size());
    if (change || newCurrent != 0) {
        setCurrent(newCurrent);
    }
    adjustPosition();
    sig_moved.raise(getId());
    drawWidget();
}

void
WFleetMemberList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    // Allocate space: 1 em for icon, 5 em for fcode, remainder for name
    int space = getExtent().w;
    int iconWidth = std::min(space, font_ems[FONT_NORMAL]);
    space -= iconWidth;
    int fcodeWidth = std::min(space, 3*font_ems[FONT_NORMAL]);
    space -= fcodeWidth;
    int nameWidth = space;

    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        int slot = getTop() + i;
        if (slot < int(content.size())) {
            // Show a member
            uint16_t value = content[slot];
            GShip& mem = getDisplayedTurn().getCurrentUniverse().getShip(value & fIdMask);

            // Icon. PCC 1.x has bullet for leader, 't' for tower, dot for away, dim color for towee
            const char* icon;
            if (value & fLeader) {
                icon = UTF_SQUARE_BULLET;
            } else if (value & fTower) {
                icon = "t";
            } else if (value & fAway) {
                icon = UTF_MIDDLE_DOT;
            } else {
                icon = "";
            }
            if (value & fTowed) {
                ctx.setColor(UIColor::tc_Blue);
            }

            // Draw it
            ctx.setTextAlign(1, 0);
            outText(ctx, area.x + iconWidth/2, area.y, icon);
            ctx.setTextAlign(0, 0);
            outTextF(ctx, area.x + iconWidth, area.y, nameWidth, format("%d: %s", mem.getId(), mem.getName(GObject::PlainName)));
            outTextF(ctx, area.x + iconWidth + nameWidth, area.y, fcodeWidth, mem.getFCode());

            // Selection icon
            if (mem.isMarked()) {
                ctx.setColor(UIColor::tc_Selection);
                drawSelection(ctx, GfxPoint(area.x + iconWidth/2, area.y + area.h/2-1), 1, 2);
            }
        }
    }
}

int
WFleetMemberList::getCurrentMemberId() const
{
    if (getCurrent() < int(content.size())) {
        return content[getCurrent()] & fIdMask;
    } else {
        return 0;
    }
}

/** Check whether any displayed member is dirty. */
bool
WFleetMemberList::isAnyMemberDirty() const
{
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    for (std::vector<uint16_t>::size_type i = 0; i < content.size(); ++i) {
        if (univ.getShip(content[i] & fIdMask).isDirty()) {
            return true;
        }
    }
    return false;
}
