/**
  *  \file game/lock.cc
  *  \brief Locking onto Objects
  *
  *  This module provides functions for locking at objects (mouse click
  *  into starchart). In addition, it provides a means for configuring
  *  what mouse button locks onto what.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "game/lock.h"
#include "game/univ.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/mine.h"
#include "game/drawing.h"
#include "game/pref.h"

/** \class GLockData
    \brief Picking the nearest point of a set

    This class encapsulates the logic for picking the nearest point of
    a set. Instanciate it with the right parameters, invoke addPoint() for
    all points, and grab the result using getPoint().

    This class automatically handles wrap. */

/** Unbounded area.
    \param clicked point the user clicked on
    \param marked  true to only process marked points */
GLockData::GLockData(GPoint clicked, bool marked)
    : clicked(clicked), min(INT_MIN, INT_MIN), max(INT_MAX, INT_MAX),
      found(clicked), found_obj(0), marked(marked), min_dist(LONG_MAX)
{ }

/** Bounded area.
    \param clicked point the user clicked on
    \param min     minimum coordinates
    \param max     maximum coordinates
    \param marked  true to only process marked points */
GLockData::GLockData(GPoint clicked, GPoint min, GPoint max, bool marked)
    : clicked(clicked), min(min), max(max), found(clicked), found_obj(0), marked(marked), min_dist(LONG_MAX)
{ }

/** Check point for inclusion in result. Does not mangle the point any
    further, just checks it. */
void
GLockData::addPointRaw(GPoint pt, const GMapObject* obj)
{
    if (pt.x >= min.x && pt.y >= min.y && pt.x <= max.x && pt.y <= max.y) {
        long dist = distanceSquared(pt, clicked);
        if (dist < min_dist) {
            found     = pt;
            found_obj = obj;
            min_dist  = dist;
        }
    }
}

/** Add a point to the set.
    \param pt      coordinates
    \param marked  true iff the object at these coordinates is marked */
void
GLockData::addPoint(GPoint pt, bool marked, const GMapObject* obj)
{
    if (this->marked && !marked)
        return;

    switch (chart_conf.getMode()) {
     case GChartConfiguration::Flat:
     case GChartConfiguration::Wrapped:
        // Fast version
        addPointRaw(pt.getSimpleNearestAlias(clicked), obj);
        break;

     case GChartConfiguration::Circular:
        // Generic version (would work for Flat/Wrapped as well but be slower)
        for (int i = 0, n = chart_conf.getNumPointImages(); i < n; ++i) {
            GPoint tmp;
            if (chart_conf.getPointAlias(pt, tmp, i, true)) {
                addPointRaw(tmp, obj);
            }
        }
        break;
    }
}

/** Add an object to the set. This uses the object's coordinates and selection status.
    \param obj the object */
void
GLockData::addObject(const GMapObject& obj)
{
    addPoint(obj.getPos(), obj.isMarked(), &obj);
}

/** Lock onto planets. Processes all planets in /univ/ using the lock
    data /data/. */
void
findPlanet(const GUniverse& univ, GLockData& data)
{
    const GAnyPlanetType& ty = univ.ty_any_planets;
    for (int pid = ty.findNextIndex(0); pid != 0; pid = ty.findNextIndex(pid)) {
        data.addObject(univ.getPlanet(pid));
    }
}

/** Lock onto ships. Processes all ships in /univ/ using the lock data
    /data/. */
void
findShip(const GUniverse& univ, GLockData& data)
{
    const GAnyShipType& ty = univ.ty_any_ships;
    for (int sid = ty.findNextIndex(0); sid != 0; sid = ty.findNextIndex(sid)) {
        data.addObject(univ.getShip(sid));
    }
}

/** Lock onto Ufo. Processes all ufos in /univ/ using the lock data
    /data/. */
void
findUfo(const GUniverse& univ, GLockData& data)
{
    const GUfoType& uc = univ.ty_ufos;
    for (GUfoType::const_iterator i = uc.begin(); i != uc.end(); ++i)
        if (i->isExistant())
            data.addObject(*i);
}

/** Lock onto marker. Processes all markers (=drawings of type MarkerDrawing)
    from /univ/ using lock data /data/. */
void
findMarker(const GUniverse& univ, GLockData& data)
{
    const GDrawingContainer& d = univ.getDrawings();
    for (GDrawingContainer::const_iterator i = d.begin(); i != d.end(); ++i)
        if ((*i)->isVisible() && (*i)->getType() == MarkerDrawing)
            data.addPoint((*i)->getPos(), false);

    const GExplosionContainer& e = univ.getExplosions();
    for (GExplosionContainer::const_iterator i = e.begin(); i != e.end(); ++i)
        data.addPoint(i->getPos(), false);
}

/** Lock onto minefields. Processes all minefields in /univ/ using the
    lock data /data/. */
void
findMinefield(const GUniverse& univ, GLockData& data)
{
    const GMinefieldType& mines = univ.ty_minefields;
    GMinefieldType::const_iterator e = mines.end();
    for (GMinefieldType::const_iterator i = mines.begin(); i != e; ++i) {
        data.addObject(*i);
    }
}

/*************************** User Configuration **************************/

static int32_t lock_config[lm_MAX] = {
    li_Planet | li_Minefield | li_Ufo,
    li_Ship | li_Marker
};

static ValueBitsetParser parse_lockopts("planet,ship,ufo,marker,minefield");
static ConfigIntRefOption opt_lock_left (getUserPreferences(), "Lock.Left",  parse_lockopts, lock_config[lm_Left]);
static ConfigIntRefOption opt_lock_right(getUserPreferences(), "Lock.Right", parse_lockopts, lock_config[lm_Right]);

/** Configure locking. This assocates an event with a set of items to
    lock on, such as "left mouse button locks onto planets and
    Ufos".
    \param mode   Event
    \param items  Item set
    \see findItem() */
void
setLockMode(TLockMode mode, int items)
{
    lock_config[mode] = items;
}

/** Query lock mode. Returns the set of items associated with the
    specified event /mode/.
    \see setLockMode(), findItem() */
int
getLockMode(TLockMode mode)
{
    return lock_config[mode];
}

/** Main entry point: do locking according to mode.
    \param univ   universe to process
    \param data   GLockData object that does the work
    \param mode   event, corresponds to the mouse button the user pressed
    \see setLockMode() */
void
findItem(const GUniverse& univ, GLockData& data, TLockMode mode)
{
    int items = lock_config[mode];
    if (items & li_Planet)
        findPlanet(univ, data);
    if (items & li_Ship)
        findShip(univ, data);
    if (items & li_Ufo)
        findUfo(univ, data);
    if (items & li_Marker)
        findMarker(univ, data);
    if (items & li_Minefield)
        findMinefield(univ, data);
}

/** Get config option corresponding to a lock mode. */
ConfigOption&
getLockModeConfigOption(TLockMode mode)
{
    return mode == lm_Left ? opt_lock_left : opt_lock_right;
}
