/**
  *  \file int/dvload.cc
  *  \brief Interpreter: Data and Value Loading
  *
  *  This module contains routines that actually belong into
  *  int/data.cc and int/value.cc, separated to reduce executable
  *  size.
  *
  *  IntValue::load references constructors of all serializable
  *  IntValue descendants. Putting it separate allows for unused
  *  classes to be optimized out during linking when load isn't
  *  actually called.
  *
  *  Also see int/vmioinst.cc.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/data.h"
#include "int/value.h"
#include "cpluslib/buffer.h"
#include "int/context.h"
#include "int/contextio.h"
#include "int/vmio.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "int/structure.h"
#include "int/subr.h"

static IntValue*
failInvalidValue(Stream& aux)
{
    throw FileFormatException(aux, _("Invalid value in data segment; file probably written by newer version of PCC"));
}

IntValue*
IntValue::load(const IntTagNode& sv, Stream& aux)
{
    switch (sv.tag) {
     case IntTagNode::Tag_Empty:
        return 0;

     case IntTagNode::Tag_Integer:
        return new IntIntIntValue(sv.value);

     case IntTagNode::Tag_Boolean:
        return new IntIntBoolValue(sv.value);

     case IntTagNode::Tag_String:
     case IntTagNode::Tag_LongString:
        return new IntStringValue(sv, aux);

     case IntTagNode::Tag_32bitFP:
     case IntTagNode::Tag_FPZero:
        return new IntFloatValue(sv, aux);

     case IntTagNode::Tag_Blob:
        return new IntBlobValue(sv, aux);

     case IntTagNode::Tag_FileHandle:
        return new IntFileValue(sv, aux);

     case IntTagNode::Tag_BCO:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntSubroutineValue(vlc->getBCO(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     case IntTagNode::Tag_Array:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntArray(vlc->getArray(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     case IntTagNode::Tag_Hash:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntHash(vlc->getHash(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     case IntTagNode::Tag_Struct:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntStructureValue(vlc->getStructureValue(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     case IntTagNode::Tag_StructType:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntStructureType(vlc->getStructureType(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     default:
        if ((sv.tag & 0x00FF) == 0) {
            IntContext* cv = loadContext(sv, aux);
            if (!cv)
                return failInvalidValue(aux);
            return cv;
        } else {
            return new IntFloatValue(sv, aux);
        }
    }
}


/** Load from stream.
    \param in Stream to load from
    \param slots Number of slots to load (slots with higher numbers will not be touched) */
void
IntDataSegment::load(Stream& in, uint32_t slots)
{
    uint32_t headerSize = slots * IntTagNode::size;
    Array<char> headers(headerSize);

    // Read headers
    in.readT(headers.ptr(), headerSize);

    // Read elements
    for (uint32_t i = 0; i < slots; ++i) {
        // Read tag node
        IntTagNode sv;
        getStructure(headers.ptr() + (IntTagNode::size * i), sv);

        // Generate node
        setNew(i, IntValue::load(sv, in));
    }
}
