/**
  *  \file int/if/mineif.cc
  *  \brief Interpreter Game Interface: Minefields
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/mineif.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/mine.h"
#include "int/compound.h"
#include "int/data.h"
#include "int/if/ifutil.h"
#include "int/if/playerif.h"
#include "int/interror.h"
#include "int/value.h"
#include "int/args.h"
#include "cpluslib/format.h"
#include "int/contextio.h"
#include "int/if/objif.h"
#include "int/propenum.h"

/* @q Minefield(id:Int):Obj (Function, Context)
   Access minefield properties.
   Use as
   | ForEach Minefield Do ...
   or
   | With Minefield(n) Do ...

   @diff This function was available for use in %With under the name %Minefields() since PCC 1.0.11.
   Do not use the name %Minefields in new code, it is not supported by PCC2; use %Minefield instead.

   @see int:index:group:minefieldproperty|Minefield Properties, int:index:group:minefieldcommand|Minefield Commands
   @since PCC 1.0.18, PCC2 1.99.8 */
IntValue*
IFMinefieldGet(IntArgBlock& args)
{
    int32_t mid;

    args.checkArgumentCount(1);
    if (!checkIntArg(mid, args.getNext()))
        return 0;

    if (!haveDisplayedTurn())
        return 0;

    return new IntMinefieldContext(mid);
}

IntContext*
IFMinefieldMake()
{
    if (!haveDisplayedTurn())
        return 0;

    int mid = current_minefield_type.findNextIndex(0);
    if (mid != 0)
        return new IntMinefieldContext(mid);
    else
        return 0;
}

int32_t
IFMinefieldDim(int32_t dim)
{
    return dim == 0 ? 1 : 10001;
}

/*************************** Minefield Methods ***************************/

/* @q Delete (Minefield Command)
   Deletes the current minefield.
   Like <kbd>Del</kbd> in the minefield window,
   this can be used to delete minefields which are known to be out-of-date:
   | ForEach Minefield Do
   |   If Owner$=My.Race$ And Scanned<>3 Then Delete
   | Next
   deletes all your minefields which were not scanned this turn.
   If you're getting Winplan RSTs, you scan all your minefields each turn,
   so those you do not scan do no longer exist and will be deleted by the above command.

   After this command, all properties of the current minefield will yield EMPTY.
   @since PCC 1.0.12, PCC2 1.99.17 */
static void
IFMinefieldDelete(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // No arguments
    args.checkArgumentCount(0);

    // Delete it
    univ.ty_minefields.erase(obj.getId());
}

/************************** IntMinefieldContext **************************/

namespace {
    enum { MinefieldProperty, MinefieldMethod, OwnerProperty };
}

static const TIntMapping minefield_mapping[] = {
    { "DELETE",     2,            MinefieldMethod,   thProcedure },
    { "ID",         impId,        MinefieldProperty, thInt },
    { "LASTSCAN",   impLastScan,  MinefieldProperty, thInt },
    { "LOC.X",      impLocX,      MinefieldProperty, thInt },
    { "LOC.Y",      impLocY,      MinefieldProperty, thInt },
    { "MARK",       0,            MinefieldMethod,   thProcedure },
    { "MARKED",     impMarked,    MinefieldProperty, thBool },
    { "OWNER",      iplShortName, OwnerProperty,     thString },
    { "OWNER$",     iplId,        OwnerProperty,     thInt },
    { "OWNER.ADJ",  iplAdjName,   OwnerProperty,     thString },
    { "RADIUS",     impRadius,    MinefieldProperty, thInt },
    { "SCANNED",    impScanType,  MinefieldProperty, thInt },
    { "TYPE",       impTypeStr,   MinefieldProperty, thString },
    { "TYPE$",      impTypeCode,  MinefieldProperty, thInt },
    { "UNITS",      impUnits,     MinefieldProperty, thInt },
    { "UNMARK",     1,            MinefieldMethod,   thProcedure },
};

static const IntObjectProcedureValue::call_t minefield_methods[] = {
    IFObjMark,                  // 0
    IFObjUnmark,                // 1
    IFMinefieldDelete,          // 2
};

/** Constructor.
    \param mid Minefield Id */
IntMinefieldContext::IntMinefieldContext(int mid)
    : IntObjectContext(current_minefield_type, mid)
{ }

/** Destructor. */
IntMinefieldContext::~IntMinefieldContext()
{ }

int32_t
IntMinefieldContext::lookup(const string_t& name)
{
    return lookupName(name, minefield_mapping, countof(minefield_mapping));
}

void
IntMinefieldContext::set(int32_t index, IntValue* value)
{
    GMinefield* mf = dynamic_cast<GMinefield*>(getObject());
    if (!mf)
        throw IntError::notAssignable();
    switch (minefield_mapping[index].domain) {
     case MinefieldProperty:
        setMinefieldProperty(*mf, IntMinefieldProperty(minefield_mapping[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntMinefieldContext::get(int32_t index)
{
    GMinefield* mf = dynamic_cast<GMinefield*>(getObject());
    if (mf == 0) {
        return 0;
    }
    switch (minefield_mapping[index].domain) {
     case MinefieldProperty:
        return getMinefieldProperty(*mf, IntMinefieldProperty(minefield_mapping[index].index));
     case OwnerProperty:
        return getPlayerProperty(mf->getOwner(), IntPlayerProperty(minefield_mapping[index].index));
     case MinefieldMethod:
        return new IntObjectProcedureValue(*mf, *getCurrentUniverse(), minefield_methods[minefield_mapping[index].index]);
    }
    return 0;
}

IntContext*
IntMinefieldContext::clone()
{
    return new IntMinefieldContext(getCurrentIndex());
}

string_t
IntMinefieldContext::toString(bool /*readable*/)
{
    return format("Minefield(%d)", getCurrentIndex());
}

void
IntMinefieldContext::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = Tag_Minefield;
    sv.value = getCurrentIndex();
}

void
IntMinefieldContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(minefield_mapping, countof(minefield_mapping));
}

/** Get minefield property.
    \param mf Minefield to query
    \param imp Property to query
    \return property value, newly allocated */
IntValue*
getMinefieldProperty(const GMinefield& mf, IntMinefieldProperty imp)
{
    switch (imp) {
     case impId:
        /* @q Id:Int (Minefield Property)
           Id of this minefield. */
        return makeIntValue(mf.getId());
     case impLastScan:
        /* @q LastScan:Int (Minefield Property)
           Turn when minefield was last scanned. */
        return makeIntValue(mf.getTurnLastSeen());
     case impLocX:
        /* @q Loc.X:Int (Minefield Property)
           X location of center of minefield. */
        return makeIntValue(mf.getPos().x);
     case impLocY:
        /* @q Loc.Y:Int (Minefield Property)
           Y location of center of minefield. */
        return makeIntValue(mf.getPos().y);
     case impMarked:
        /* @q Marked:Bool (Minefield Property)
           True if minefield is marked. */
        return makeBoolValue(mf.isMarked());
     case impRadius:
        /* @q Radius:Int (Minefield Property)
           Minefield radius in ly. */
        return makeIntValue(mf.getRadius());
     case impScanType:
        /* @q Scanned:Int (Minefield Property)
           Last reported action on this minefield.
           <table>
            <tr><td width="1">0</td><td width="20">Not scanned this turn</td></tr>
            <tr><td width="1">1</td><td width="20">Laid this turn</td></tr>
            <tr><td width="1">2</td><td width="20">Swept this turn</td></tr>
            <tr><td width="1">3</td><td width="20">Scanned this turn</td></tr>
           </table> */
        // FIXME: PCC 1.x checks for <0. Can this happen?
        return makeIntValue(mf.getAction());
     case impTypeCode:
        /* @q Type$:Bool (Minefield Property)
           True if this is a web mine field. */
        return makeBoolValue(mf.isWeb());
     case impTypeStr:
        /* @q Type:Str (Minefield Property)
           Minefield type, human-readable.
           One of "Web Mines" or "Mines". */
        return makeStringValue(mf.isWeb() ? "Web Mines" : "Mines");
     case impUnits:
        /* @q Units:Int (Minefield Property)
           Number of mine units. */
        return makeIntValue(mf.getUnits());
    }
    return 0;
}

/** Set minefield property. */
void
setMinefieldProperty(GMinefield& /*mf*/, IntMinefieldProperty /*imp*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}
