/**
  *  \file io/memdir.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "io/memdir.h"
#include "util/except.h"
#include "util/nls.h"

static string_t
buildMemoryStreamName(string_t dirname, string_t basename)
{
    return dirname + "(" + basename + ")";
}

/************************* MemoryDirectory::Entry ************************/

class MemoryDirectory::Entry : public DirectoryEntry {
    Ptr<MemoryDirectory> dir;
    string_t basename;

 public:
    Entry(Ptr<MemoryDirectory> dir, string_t basename);
    ~Entry();

    void           getInfo();
    Ptr<Directory> openDir();
    Ptr<Stream>    openFile(int mode);
    Ptr<Directory> openParent();
    void           erase();
    void           renameTo(string_t name);
    void           createAsDirectory();
    string_t       getBaseName() const;
};

MemoryDirectory::Entry::Entry(Ptr<MemoryDirectory> dir, string_t basename)
    : DirectoryEntry(buildMemoryStreamName(dir->getName(), basename)),
      dir(dir),
      basename(basename)
{ }

MemoryDirectory::Entry::~Entry()
{ }

void
MemoryDirectory::Entry::getInfo()
{
    Ptr<MemoryStream> str = dir->getStreamByName(basename, false);
    if (str)
        setInfo(str->getSize(), FT_FILE);
}

Ptr<Directory>
MemoryDirectory::Entry::openDir()
{
    throw FileOpenException(name, _("No such directory"));
}

Ptr<Stream>
MemoryDirectory::Entry::openFile(int mode)
{
    Ptr<MemoryStream> str = dir->getStreamByName(basename, (mode & Stream::C_CREATE) != 0);
    if (!str)
        throw FileOpenException(name, _("No such file"));

    if (mode & Stream::C_CREATE) {
        // truncate it if we're creating it
        Buffer<char> tmp;
        str->swapBuffer(tmp);
    }

    Ptr<Stream> result = str->dup();
    result->seek(0);
    return result;
}

Ptr<Directory>
MemoryDirectory::Entry::openParent()
{
    return dir;
}

void
MemoryDirectory::Entry::erase()
{
    if (!dir->eraseStream(basename))
        throw FileIOException(name, _("No such file"));
    clearInfo();
}

void
MemoryDirectory::Entry::renameTo(string_t name)
{
    if (!dir->renameStream(basename, name))
        throw FileIOException(this->name, _("No such file"));
    clearInfo();
}

void
MemoryDirectory::Entry::createAsDirectory()
{
    throw FileOpenException(getName(), _("Cannot create directory"));
}

string_t
MemoryDirectory::Entry::getBaseName() const
{
    return basename;
}

/************************* MemoryDirectory::Enum *************************/

class MemoryDirectory::Enum : public Enumeration<Ptr<DirectoryEntry> > {
    Ptr<MemoryDirectory> dir;
    MemoryDirectory::container_t::iterator it;
 public:
    Enum(Ptr<MemoryDirectory> dir);

    bool hasMoreElements();
    Ptr<DirectoryEntry> getNextElement();
};

MemoryDirectory::Enum::Enum(Ptr<MemoryDirectory> dir)
    : dir(dir),
      it(dir->content.begin())
{
    // FIXME: to make this robust, we'd have to deal with parallel
    // erase/rename requests for our current item
}

bool
MemoryDirectory::Enum::hasMoreElements()
{
    return it != dir->content.end();
}

Ptr<DirectoryEntry>
MemoryDirectory::Enum::getNextElement()
{
    return new Entry(dir, (it++)->first);
}

/**************************** MemoryDirectory ****************************/

MemoryDirectory::MemoryDirectory(string_t name)
    : name(name)
{ }

Ptr<MemoryDirectory>
MemoryDirectory::create(string_t name)
{
    return new MemoryDirectory(name);
}

MemoryDirectory::~MemoryDirectory()
{ }

/* Directory methods */
Ptr<DirectoryEntry>
MemoryDirectory::getEntryByName(string_t name)
{
    return new Entry(this, name);
}

Ptr<Enumeration<Ptr<DirectoryEntry> > >
MemoryDirectory::getEntries()
{
    return new Enum(this);
}

Ptr<Directory>
MemoryDirectory::getParent()
{
    return 0;
}

string_t
MemoryDirectory::getName()
{
    return name;
}

/** Get stream by name. Stream will be created if requested.
    \param name basename
    \param create_flag true to create stream if it does not exist, false to return 0 in this case */
Ptr<MemoryStream>
MemoryDirectory::getStreamByName(string_t name, bool create_flag)
{
    container_t::iterator i = content.find(name);
    if (i != content.end()) {
        return i->second;
    } else if (create_flag) {
        Ptr<MemoryStream> s = new MemoryStream();
        content[name] = s;
        s->setName(buildMemoryStreamName(getName(), name));
        return s;
    } else {
        return 0;
    }
}

/** Erase stream by name.
    \param name Basename
    \return true if erased, false if did not exist */
bool
MemoryDirectory::eraseStream(string_t name)
{
    container_t::iterator i = content.find(name);
    if (i != content.end()) {
        content.erase(i);
        return true;
    } else {
        return false;
    }
}

/** Rename stream by name.
    \param oldname old basename
    \param newname new basename
    \return true if successfully renamed, false if oldname does not exist */
bool
MemoryDirectory::renameStream(string_t oldname, string_t newname)
{
    container_t::iterator oi = content.find(oldname);
    container_t::iterator ni = content.find(newname);
    if (oi == ni) {
        // Same. Error if both are end()
        return oi != content.end();
    } else {
        // Different. Check preconditions.
        if (oi == content.end())
            // Not found
            return false;

        // We can rename
        content[newname] = oi->second;
        oi->second->setName(buildMemoryStreamName(getName(), newname));
        content.erase(oi);
        return true;
    }
}
